import time

import iridauploader.config as config


class DirectoryStatus:
    """
    When looking for runs in a directory (or list of directories) this object is used to
    contain the relevant data.
    """

    # States that are valid for the status field

    # New runs, ready to upload
    NEW = 'new'
    # Delayed runs are new runs that have been discovered as new, but must wait before upload
    # When a run is found with the Delayed status, The uploader will check if enough time has passed to start the upload
    DELAYED = 'delayed'
    # Invalid is used when a run directory does not have the base requirements to act as a run
    # Never written to a status file
    INVALID = 'invalid'
    # Parsing/Upload has started/partially completed for this run
    PARTIAL = 'partial'
    # Parsing/Upload has stopped because of some error.
    ERROR = 'error'
    # Parsing/Upload has completed successfully
    COMPLETE = 'complete'

    VALID_STATUS_LIST = [
        NEW,
        DELAYED,
        INVALID,
        PARTIAL,
        ERROR,
        COMPLETE
    ]

    # Status field for a sequencing run
    JSON_STATUS_FIELD = "Upload Status"
    JSON_DIRECTORY_FIELD = "Directory"
    JSON_DATE_TIME_FIELD = "Date Time"
    JSON_DATE_TIME_FORMAT = "%Y-%m-%d %H:%M"
    JSON_RUN_ID_FIELD = "Run ID"
    JSON_IRIDA_INSTANCE_FIELD = "IRIDA Instance"
    JSON_MESSAGE_FIELD = "Message"
    JSON_SAMPLES_UPLOADED_FIELD = "Sample Status"

    def __init__(self, directory, status=NEW, message=None):
        """
        :param directory: Directory of a potential run
        :param status: Default = NEW, must be one of VALID_STATUS_LIST
        :param message: Text to be added to the status file
        """
        self._directory = directory
        # use the status validation method
        self.status = status
        self._message = message
        self._run_id = None
        self._sample_status_list = None
        self._time = None
        self._irida_instance = None

    def init_file_status_list_from_sequencing_run(self, sequencing_run):
        """
        Creates list of SampleStatus objects (project/sample pairs w/ upload status) and sets to uploaded=False
        :param sequencing_run:
        :return:
        """
        self._sample_status_list = []
        for project in sequencing_run.project_list:
            for sample in project.sample_list:
                self._sample_status_list.append(self.SampleStatus(sample.sample_name, project.id))

    def set_sample_uploaded(self, sample_name, project_id, uploaded):
        """
        Finds the SampleStatus object in sample_status_list and updates uploaded to True
        :param sample_name:
        :param project_id:
        :param uploaded:
        :return:
        """
        for status in self._sample_status_list:
            if status.equals(sample_name=sample_name, project_id=project_id):
                status.uploaded = uploaded

    def sample_status_to_dict(self):
        """
        Creates list of dicts that can be easily written as json
        :return:
        """
        if self._sample_status_list is None:
            return None
        else:
            return [o.to_dict() for o in self._sample_status_list]

    @property
    def directory(self):
        return self._directory

    @directory.setter
    def directory(self, directory):
        self._directory = directory

    @property
    def run_id(self):
        return self._run_id

    @run_id.setter
    def run_id(self, run_id):
        self._run_id = run_id

    @property
    def status(self):
        """
        status property, setter is overridden by @status.setter
        :return: status
        """
        return self._status

    @status.setter
    def status(self, status):
        """
        Status setter
        Only accepts status's from the predefined VALID_STATUS_LIST
        :param status: Status to set on object
        :return:
        """
        if status not in DirectoryStatus.VALID_STATUS_LIST:
            raise Exception("Invalid status written to DirectoryStatus object.")
        self._status = status

    def status_equals(self, status):
        """
        Checks if status on object is the same as status given.
        Only accepts status's from the predefined VALID_STATUS_LIST
        :param status: check given status against current status
        :return:
        """
        if status not in DirectoryStatus.VALID_STATUS_LIST:
            raise Exception("Invalid status '{}' used for comparison.".format(status))
        return self._status == status

    @property
    def message(self):
        return self._message

    @message.setter
    def message(self, message):
        self._message = message

    @property
    def time(self):
        if self._time is None:
            return None
        return self._time

    @time.setter
    def time(self, formatted_time):
        if formatted_time is None:
            self._time = None
        elif type(formatted_time) is not str:
            raise TypeError("Time given is not a string: {}".format(formatted_time))
        else:
            self._time = time.strptime(formatted_time, self.JSON_DATE_TIME_FORMAT)

    @property
    def irida_instance(self):
        if self._irida_instance is None:
            self._irida_instance = config.read_config_option('base_url')
        return self._irida_instance

    @irida_instance.setter
    def irida_instance(self, irida_instance):
        self._irida_instance = irida_instance

    def get_sample_status_list(self):
        return self._sample_status_list

    def to_json_dict(self):
        sample_status_dict = self.sample_status_to_dict()

        json_dict = {
            self.JSON_STATUS_FIELD: self.status,
            self.JSON_DIRECTORY_FIELD: self.directory,
            self.JSON_MESSAGE_FIELD: self.message,
            self.JSON_DATE_TIME_FIELD: time.strftime(self.JSON_DATE_TIME_FORMAT),  # Generate a new timestamp
            self.JSON_RUN_ID_FIELD: self.run_id if self.run_id is not None else "",
            self.JSON_IRIDA_INSTANCE_FIELD: self.irida_instance,
            self.JSON_SAMPLES_UPLOADED_FIELD: sample_status_dict if sample_status_dict is not None else "",
        }

        return json_dict

    @staticmethod
    def init_from_json_dict(json_dict, directory=None):
        if directory is not None:
            new_directory_status = DirectoryStatus(directory)
        else:
            new_directory_status = DirectoryStatus(DirectoryStatus._get_field_or_none(
                json_dict, DirectoryStatus.JSON_DIRECTORY_FIELD))
        new_directory_status.status = DirectoryStatus._get_field_or_none(
            json_dict, DirectoryStatus.JSON_STATUS_FIELD)
        new_directory_status.message = DirectoryStatus._get_field_or_none(
            json_dict, DirectoryStatus.JSON_MESSAGE_FIELD)
        new_directory_status.time = DirectoryStatus._get_field_or_none(
            json_dict, DirectoryStatus.JSON_DATE_TIME_FIELD)
        new_directory_status.run_id = DirectoryStatus._get_field_or_none(
            json_dict, DirectoryStatus.JSON_RUN_ID_FIELD)
        new_directory_status.irida_instance = DirectoryStatus._get_field_or_none(
            json_dict, DirectoryStatus.JSON_IRIDA_INSTANCE_FIELD)

        new_directory_status._sample_status_list = DirectoryStatus._get_sample_status_list_from_json(
            json_dict, DirectoryStatus.JSON_SAMPLES_UPLOADED_FIELD)

        return new_directory_status

    @staticmethod
    def _get_field_or_none(json_dict, field):
        return json_dict[field] if field in json_dict else None

    @staticmethod
    def _get_sample_status_list_from_json(json_dict, samples_field):
        new_sample_status_list = []

        samples_uploaded_field = DirectoryStatus._get_field_or_none(
            json_dict, samples_field)

        if samples_uploaded_field is not None:
            for sample_dict in json_dict[samples_field]:
                new_sample_status_list.append(DirectoryStatus.SampleStatus(
                    sample_name=sample_dict[DirectoryStatus.SampleStatus.SAMPLE_NAME_FIELD],
                    project_id=sample_dict[DirectoryStatus.SampleStatus.PROJECT_ID_FIELD],
                    uploaded=sample_dict[DirectoryStatus.SampleStatus.UPLOADER_FIELD],
                ))

        return new_sample_status_list

    class SampleStatus:
        """
        Contains a sample name, project id and upload status
        This is used to simply jsonify the directory status info
        """

        SAMPLE_NAME_FIELD = "Sample Name"
        PROJECT_ID_FIELD = "Project ID"
        UPLOADER_FIELD = "Uploaded"

        def __init__(self, sample_name, project_id, uploaded=False):
            """
            Init SampleStatus Object, initializes uploaded to False

            :param sample_name: sample.sample_name from Sample object
            :param project_id: project.id from Project object
            """
            self._sample_name = sample_name
            self._project_id = project_id
            # if `uploaded` is a string, convert it to a boolean
            if type(uploaded) is str:
                uploaded = uploaded.lower() in ['true', '1', 't', 'y', 'yes']
            self._uploaded = uploaded

        @property
        def uploaded(self):
            return self._uploaded

        @uploaded.setter
        def uploaded(self, uploaded):
            self._uploaded = uploaded

        @property
        def sample_name(self):
            return self._sample_name

        @property
        def project_id(self):
            return self._project_id

        def equals(self, sample_name, project_id):
            """
            Verifies sample id and project id match current instance.
            Upload status bool is not inspected
            :param sample_name:
            :param project_id:
            :return:
            """
            return self.sample_name == sample_name and self.project_id == project_id

        def to_dict(self):
            """
            Creates and returns a json ready dict that gets used when updating the status file
            :return:
            """
            return {
                self.SAMPLE_NAME_FIELD: self.sample_name,
                self.PROJECT_ID_FIELD: self.project_id,
                self.UPLOADER_FIELD: str(self.uploaded)
            }



import unittest
from unittest.mock import patch
from os import path

import iridauploader.parsers as parsers
from iridauploader.core import parsing_handler

path_to_module = path.abspath(path.dirname(__file__))
if len(path_to_module) == 0:
    path_to_module = '.'


class TestGetParserFromConfig(unittest.TestCase):
    """
    When a new parser is added to the uploader, a test that it can be gotten should be added here
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    @patch("iridauploader.core.parsing_handler.config.read_config_option")
    def test_get_miseq_parser(self, mock_read_config_option):
        # force the handler to grab 'miseq' when it tries to call config
        mock_read_config_option.side_effect = ["miseq"]

        res = parsing_handler.get_parser_from_config()
        # verify we grabbed the right parser
        self.assertEqual(type(res), parsers.parsers.miseq.Parser)

    @patch("iridauploader.core.parsing_handler.config.read_config_option")
    def test_get_directory_parser(self, mock_read_config_option):
        # force the handler to grab 'directory' when it tries to call config
        mock_read_config_option.side_effect = ["directory"]

        res = parsing_handler.get_parser_from_config()
        # verify we grabbed the right parser
        self.assertEqual(type(res), parsers.parsers.directory.Parser)


class TestParseAndValidate(unittest.TestCase):
    """
    Tests the core.parsing_handler.parse_and_validate function
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    @patch("iridauploader.core.uniform_file_count_validator.validate_uniform_file_count")
    @patch("iridauploader.core.file_size_validator.validate_file_size_minimum")
    @patch("iridauploader.core.parsing_handler.model_validator.validate_sequencing_run")
    @patch("iridauploader.core.parsing_handler.get_parser_from_config")
    def test_all_functions_called(
            self,
            mock_get_parser,
            mock_validate_model,
            mock_validate_file_size,
            mock_validate_uniform_file_count):
        """
        Makes sure that all relevant functions are called so that it will parse and validate fully
        :return:
        """
        # add mock data (as strings) to the function calls that are essential
        mock_parser_instance = unittest.mock.MagicMock()
        mock_parser_instance.get_sample_sheet.side_effect = ["mock_sample_sheet"]
        mock_parser_instance.get_sequencing_run.side_effect = ["mock_sequencing_run"]

        mock_get_parser.side_effect = [mock_parser_instance]

        mock_validation_result = unittest.mock.MagicMock()
        mock_validation_result.is_valid.side_effect = [True, True, True]

        mock_validate_model.side_effect = [mock_validation_result]
        mock_validate_file_size.side_effect = [mock_validation_result]
        mock_validate_uniform_file_count.side_effect = [mock_validation_result]

        res = parsing_handler.parse_and_validate("mock_directory")

        # verify that the functions were called the mock data we set up
        mock_parser_instance.get_sample_sheet.assert_called_once_with("mock_directory")
        mock_parser_instance.get_sequencing_run.assert_called_once_with("mock_sample_sheet")
        mock_validate_model.assert_called_once_with("mock_sequencing_run")
        mock_validate_file_size.assert_called_once_with("mock_sequencing_run")
        self.assertEqual(res, "mock_sequencing_run")

    @patch("iridauploader.core.parsing_handler.model_validator.validate_sequencing_run")
    @patch("iridauploader.core.parsing_handler.get_parser_from_config")
    def test_invalid_sequencing_run(self, mock_get_parser, mock_validate):
        """
        Check that an exception is thrown when a given SequencingRun is given
        :return:
        """
        # add mock data (as strings) to the function calls that are essential
        mock_parser_instance = unittest.mock.MagicMock()
        mock_parser_instance.get_sample_sheet.side_effect = ["mock_sample_sheet"]
        mock_parser_instance.get_sequencing_run.side_effect = ["mock_sequencing_run"]

        mock_get_parser.side_effect = [mock_parser_instance]

        mock_validation_result = unittest.mock.MagicMock()
        mock_validation_result.is_valid.side_effect = [False]

        mock_validate.side_effect = [mock_validation_result]

        # make sure the invalid sequencing run raises a ValidationError
        with self.assertRaises(parsers.exceptions.ValidationError):
            parsing_handler.parse_and_validate("mock_directory")

    @patch("iridauploader.core.parsing_handler.model_validator.validate_sequencing_run")
    @patch("iridauploader.core.parsing_handler.get_parser_from_config")
    def test_invalid_sequencing_dir(self, mock_get_parser, mock_validate):
        """
        Check that an exception is thrown when a bad directory is given
        :return:
        """
        # add mock data (as strings) to the function calls that are essential
        mock_parser_instance = unittest.mock.MagicMock()
        mock_parser_instance.get_sample_sheet.side_effect = [parsers.exceptions.DirectoryError("", "")]

        mock_get_parser.side_effect = [mock_parser_instance]

        # make sure the invalid sequencing run raises a ValidationError
        with self.assertRaises(parsers.exceptions.DirectoryError):
            parsing_handler.parse_and_validate("mock_directory")

    @patch("iridauploader.core.parsing_handler.model_validator.validate_sequencing_run")
    @patch("iridauploader.core.parsing_handler.get_parser_from_config")
    def test_invalid_sample_sheet(self, mock_get_parser, mock_validate):
        """
        Check that an exception is thrown when an invalid sample sheet is given
        :return:
        """
        # add mock data (as strings) to the function calls that are essential
        mock_parser_instance = unittest.mock.MagicMock()
        mock_parser_instance.get_sample_sheet.side_effect = ["mock_sample_sheet"]
        mock_parser_instance.get_sequencing_run.side_effect = [parsers.exceptions.ValidationError("", "")]

        mock_get_parser.side_effect = [mock_parser_instance]

        mock_validation_result = unittest.mock.MagicMock()
        mock_validation_result.is_valid.side_effect = [False]

        mock_validate.side_effect = [mock_validation_result]

        # make sure the invalid sequencing run raises a ValidationError
        with self.assertRaises(parsers.exceptions.ValidationError):
            parsing_handler.parse_and_validate("mock_directory")

    @patch("iridauploader.core.uniform_file_count_validator.validate_uniform_file_count")
    @patch("iridauploader.core.file_size_validator.validate_file_size_minimum")
    @patch("iridauploader.core.parsing_handler.model_validator.validate_sequencing_run")
    @patch("iridauploader.core.parsing_handler.get_parser_from_config")
    def test_invalid_file_count(
            self,
            mock_get_parser,
            mock_validate_model,
            mock_validate_file_size,
            mock_validate_uniform_file_count):
        """
        Makes sure that all relevant functions are called so that it will parse and validate fully
        :return:
        """
        # add mock data (as strings) to the function calls that are essential
        mock_parser_instance = unittest.mock.MagicMock()
        mock_parser_instance.get_sample_sheet.side_effect = ["mock_sample_sheet"]
        mock_parser_instance.get_sequencing_run.side_effect = ["mock_sequencing_run"]

        mock_get_parser.side_effect = [mock_parser_instance]

        mock_validation_result = unittest.mock.MagicMock()
        mock_validation_result.is_valid.side_effect = [True, False, True]

        mock_validate_model.side_effect = [mock_validation_result]
        mock_validate_file_size.side_effect = [mock_validation_result]
        mock_validate_uniform_file_count.side_effect = [mock_validation_result]

        # make sure the invalid sequencing run raises a ValidationError
        with self.assertRaises(parsers.exceptions.ValidationError):
            parsing_handler.parse_and_validate("mock_directory")

        # verify that the functions were called the mock data we set up
        mock_parser_instance.get_sample_sheet.assert_called_once_with("mock_directory")
        mock_parser_instance.get_sequencing_run.assert_called_once_with("mock_sample_sheet")
        mock_validate_model.assert_called_once_with("mock_sequencing_run")

    @patch("iridauploader.core.uniform_file_count_validator.validate_uniform_file_count")
    @patch("iridauploader.core.file_size_validator.validate_file_size_minimum")
    @patch("iridauploader.core.parsing_handler.model_validator.validate_sequencing_run")
    @patch("iridauploader.core.parsing_handler.get_parser_from_config")
    def test_invalid_file_size(
            self,
            mock_get_parser,
            mock_validate_model,
            mock_validate_file_size,
            mock_validate_uniform_file_count):
        """
        Makes sure that all relevant functions are called so that it will parse and validate fully
        :return:
        """
        # add mock data (as strings) to the function calls that are essential
        mock_parser_instance = unittest.mock.MagicMock()
        mock_parser_instance.get_sample_sheet.side_effect = ["mock_sample_sheet"]
        mock_parser_instance.get_sequencing_run.side_effect = ["mock_sequencing_run"]

        mock_get_parser.side_effect = [mock_parser_instance]

        mock_validation_result = unittest.mock.MagicMock()
        mock_validation_result.is_valid.side_effect = [True, True, False]

        mock_validate_model.side_effect = [mock_validation_result]
        mock_validate_file_size.side_effect = [mock_validation_result]
        mock_validate_uniform_file_count.side_effect = [mock_validation_result]

        # make sure the invalid sequencing run raises a ValidationError
        with self.assertRaises(parsers.exceptions.ValidationError):
            parsing_handler.parse_and_validate("mock_directory")

        # verify that the functions were called the mock data we set up
        mock_parser_instance.get_sample_sheet.assert_called_once_with("mock_directory")
        mock_parser_instance.get_sequencing_run.assert_called_once_with("mock_sample_sheet")
        mock_validate_model.assert_called_once_with("mock_sequencing_run")
        mock_validate_file_size.assert_called_once_with("mock_sequencing_run")



class BaseParser:
    """
     This class is the abstract for the different parser objects.

     Like the miseq and directory parser, a new parser class needs the following static methods
         find_single_run(directory)
         find_runs(directory)
         get_required_file_list()
         get_sample_sheet(directory)
         get_sequencing_run(sample_sheet, ...)

     """
    def __init__(self, parser_type_name, required_file_list):
        """
        Base Parser initialization
        :param parser_type_name: string: used as the run type identifier when creating the sequence run
        :param required_file_list: [file_names]
        """
        self._parser_type_name = parser_type_name
        self._required_file_list = required_file_list

    def get_required_file_list(self):
        """
        Returns a list of files that are required for a run directory to be considered valid
        :return: [files_names]
        """
        return self._required_file_list

    def get_parser_type_name(self):
        """
        Returns the parser type name, used when generating the api route to determine sequencer type
        :return:
        """
        return self._parser_type_name



class SequenceFileError(Exception):
    """An exception that's raised when errors are encountered with a sequence file.

    Examples include when files cannot be found for samples that are in the sample
    sheet, or when the server rejects a file during upload.
    """
    def __init__(self, message):
        """Initialize a SequenceFileError.

        Args:
            message: a summary message of the error.
        """
        self._message = message

    @property
    def message(self):
        return self._message

    def __str__(self):
        return self.message



import json
import logging
import os
import time

import iridauploader.config as config
from iridauploader.model.directory_status import DirectoryStatus

from . import exceptions


# Module level Constants
# These define the status files valid fields and variables

# File name
STATUS_FILE_NAME = "irida_uploader_status.info"


def get_directory_status(directory, required_file_list):
    """
    Gets the directory status based off using 'irida_uploader_status.info' files to track progress

    Returns as soon as it finds the case that applies

    :param directory: the directory to search for a run
    :param required_file_list: optional param: a list of required files that
        are required for that run to be considered valid. Example: ['SampleSheet.csv']
    :return: directory and status dictionary
    """
    # Verify directory is readable
    log_directory = config.read_config_option("log_directory")
    if not os.access(directory, os.R_OK) and not bool(log_directory):
        return DirectoryStatus(directory=directory,
                               status=DirectoryStatus.INVALID,
                               message='Directory cannot be read. Please check permissions')

    # Gets the list of files in the directory
    file_list = next(os.walk(directory))[2]

    # Legacy upload catch
    # When the irida-miseq-uploader (old uploader) ran it generated a .miseqUploaderInfo file
    # To prevent uploading runs that used this old system, we assume runs with this file are COMPLETE
    # By default they will not be picked up automatically with --batch because they are set to COMPLETE,
    # but they can still be uploaded using the --force option
    if '.miseqUploaderInfo' in file_list:
        return DirectoryStatus(directory=directory,
                               status=DirectoryStatus.COMPLETE,
                               message="Legacy uploader run. Set to complete to avoid uploading duplicate data.")

    for file_name in required_file_list:
        if file_name not in file_list:
            return DirectoryStatus(directory=directory,
                                   status=DirectoryStatus.INVALID,
                                   message='Directory is missing required file with filename {}'.format(file_name))

    # All pre-validation passed
    # Determine if status file already exists, or if the run is brand new
    if log_directory:
        status_directory = os.path.join(log_directory, directory)
    else:
        status_directory = directory
    if os.path.isfile(os.path.join(status_directory, STATUS_FILE_NAME)):  # Status file already exists, use it.
        return read_directory_status_from_file(status_directory)
    else:  # no irida_uploader_status.info file yet, has not been uploaded
        return DirectoryStatus(directory=directory, status=DirectoryStatus.NEW)


def read_directory_status_from_file(directory):
    uploader_info_file = os.path.join(directory, STATUS_FILE_NAME)

    try:
        # Read file as json
        with open(uploader_info_file, "rb") as reader:
            data = reader.read().decode()
        json_dict = json.loads(data)
        # Generate DirectoryStatus from json
        directory_status = DirectoryStatus.init_from_json_dict(json_dict, directory)
        if directory_status.status not in DirectoryStatus.VALID_STATUS_LIST:
            raise KeyError("Invalid directory status: {}".format(directory_status.status))
        # When loading from an ERROR sheet, or an old sheet, we must manually add the directory to the object
        if directory_status.directory is None:
            directory_status.directory = directory
    except KeyError as e:
        # If status file is invalid, create a new directory status with invalid and error message to return instead
        directory_status = DirectoryStatus(directory=directory, status=DirectoryStatus.INVALID, message=str(e))
    except Exception:
        # If the file cannot be read (e.g. invalid json), return invalid
        message = "Status file '{}' is malformed. Please delete this file and try again.".format(uploader_info_file)
        return DirectoryStatus(directory=directory, status=DirectoryStatus.INVALID, message=message)

    return directory_status


def write_directory_status(directory_status):
    """
    Writes a status to the status file:
    Overwrites anything that is in the file

    Writes a timestamp to the time of last written

    :param directory_status: DirectoryStatus object containing status to write to directory
    :return: None
    """
    log_directory = config.read_config_option("log_directory")
    if config.read_config_option("readonly", bool, False) is False or log_directory:
        if not os.access(directory_status.directory, os.W_OK) and not bool(log_directory):  # check if directory can be accessed, or that the log_directory is set
            raise exceptions.DirectoryError("Cannot access directory", directory_status.directory)
        elif bool(log_directory) and not os.access(log_directory, os.W_OK):  # need to check that path is filled out in addition to checking access
            raise exceptions.DirectoryError("log directory set but no write access", directory_status.directory)
        json_data = directory_status.to_json_dict()
        if log_directory:
            run_log_path = os.path.join(log_directory, os.path.basename(directory_status.directory))
            if not os.path.isdir(run_log_path):
                os.mkdir(run_log_path)
            uploader_info_file = os.path.join(run_log_path, STATUS_FILE_NAME)
        else:
            uploader_info_file = os.path.join(directory_status.directory, STATUS_FILE_NAME)
        with open(uploader_info_file, "w") as json_file:
            json.dump(json_data, json_file, indent=4, sort_keys=True)
            json_file.write("\n")


def run_is_ready_with_delay(directory_status):
    """
    Expects a NEW or DELAYED directory status

    If a NEW run is given, and the config is set to delay new runs, the run will be set to DELAYED, otherwise it's ready
    If a DELAYED run is given, the run is ready if enough time has passed, otherwise it is not ready yet.

    Writes to directory status file when set to DELAYED

    :param directory_status:
    :return: True when run is ready for upload, otherwise False
    """
    delay_minutes = config.read_config_option("delay", expected_type=int)
    logging.debug("delay_minutes is set to: " + str(delay_minutes))

    # Check if run is new, check if there's a delay
    if directory_status.status_equals(DirectoryStatus.NEW):
        if delay_minutes > 0:
            _set_run_delayed(directory_status)
            logging.info("Run has been delayed for {} minutes.".format(delay_minutes))
            run_is_ready = False
        else:
            logging.info("No delay time given for NEW run. Continuing...")
            run_is_ready = True
    # If run was delayed, check if run can now be uploaded
    elif directory_status.status_equals(DirectoryStatus.DELAYED):
        if _delayed_time_has_passed(directory_status, delay_minutes):
            logging.info("Delayed run is now ready for upload. Continuing...")
            run_is_ready = True
        else:
            logging.info("Delayed run is still not ready for upload.")
            run_is_ready = False
    # This case should be imposable
    else:
        raise Exception("Function called with invalid directory status, This should never happen.")

    return run_is_ready


def _set_run_delayed(directory_status):
    """
    Helper function to set and write directory status as delayed.

    :param directory_status:
    :return:
    """
    directory_status.status = DirectoryStatus.DELAYED
    write_directory_status(directory_status)


def _delayed_time_has_passed(directory_status, delay_minutes):
    """
    Checks if delay_minutes time has passed since directory_status.time

    See time docs for details on time modules functionality
    https://docs.python.org/3/library/time.html
    :param directory_status: time.struct_time
    :param delay_minutes: Integer
    :return: Boolean
    """
    # Delay Time is not set, run is ready for upload
    if delay_minutes == 0 or directory_status.time is None:
        return True

    # float representing the time run was found (in seconds)
    run_found_time_float = time.mktime(directory_status.time)
    # add delay time to found time
    time_plus_delay_float = run_found_time_float + (delay_minutes * 60)
    # get current time
    current_time_float = time.time()
    # compare current time to time when run is ready
    return current_time_float > time_plus_delay_float



"""
This file is used to kick off the integration tests

it can be used from command line like so:
  $ python3 start_integration_tests.py <branch>
where <branch> is the IRIDA github branch to test against
"""

import argparse
import sys

import iridauploader.tests_integration.tests_integration as tests_integration

# parse a single argument that determines which IRIDA branch we are testing against
argument_parser = argparse.ArgumentParser(description='This program parses sequencing runs and uploads them to IRIDA.')
argument_parser.add_argument('branch', help='Which IRIDA branch to run integration tests against')
argument_parser.add_argument('db_host', help="Override database host",
                             nargs='?', action='store', default=False)
argument_parser.add_argument('db_port', help="Override database port",
                             nargs='?', action='store', default=False)


def main():
    # parse argument
    args = argument_parser.parse_args()

    # Start integration tests on specified branch
    if args.db_host and args.db_port:
        return tests_integration.start(args.branch, args.db_host, args.db_port)
    else:
        return tests_integration.start(args.branch)


if __name__ == "__main__":
    sys.exit(main())



from iridauploader.parsers.exceptions.directory_error import DirectoryError
from iridauploader.parsers.exceptions.sample_sheet_error import SampleSheetError
from iridauploader.parsers.exceptions.sequence_file_error import SequenceFileError
from iridauploader.parsers.exceptions.validation_error import ValidationError
from iridauploader.parsers.exceptions.file_size_error import FileSizeError



import unittest
from unittest.mock import patch
from csv import reader
from io import StringIO

from iridauploader.parsers.directory.validation import validate_sample_sheet
from iridauploader.parsers.exceptions import SampleSheetError


class TestValidation(unittest.TestCase):
    """
    Tests valid and invalid sample sheets
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    @patch("iridauploader.parsers.common.get_csv_reader")
    def test_validate_sample_sheet_no_data_header(self, mock_csv_reader):
        """
        Given a sample sheet with no header, make sure the correct errors are included in the response
        :param mock_csv_reader:
        :return:
        """
        field_values = (
            "Sample_Name,Project_ID,File_Forward,File_Reverse\n"
            "my-sample-1, 72,file_1.fastq.gz,file_2.fastq.gz\n"
        )

        file_contents_str = (
            "{field_values}"
        ).format(field_values=field_values)

        # converts string as a pseudo file / memory file
        sample_sheet_file = StringIO(file_contents_str)

        # the call to get_csv_reader() inside parse_samples() will return
        # items inside side_effect
        mock_csv_reader.side_effect = [reader(sample_sheet_file)]

        res = validate_sample_sheet(None)

        # This should be an invalid sample sheet
        self.assertFalse(res.is_valid())
        # Only should have 2 error
        self.assertEqual(len(res.error_list), 2)
        # Error type should be SampleSheetError
        self.assertEqual(type(res.error_list[0]), SampleSheetError)
        self.assertEqual(type(res.error_list[1]), SampleSheetError)

    @patch("iridauploader.parsers.common.get_csv_reader")
    def test_validate_sample_sheet_no_data(self, mock_csv_reader):
        """
        Given a sample sheet with no data, make sure the correct errors are included in the response
        :param mock_csv_reader:
        :return:
        """
        file_contents_str = "[Data]\n"

        # converts string as a pseudo file / memory file
        sample_sheet_file = StringIO(file_contents_str)

        # the call to get_csv_reader() inside parse_samples() will return
        # items inside side_effect
        mock_csv_reader.side_effect = [reader(sample_sheet_file)]

        res = validate_sample_sheet(None)

        # This should be an invalid sample sheet
        self.assertFalse(res.is_valid())
        # Only should have 1 error
        self.assertEqual(len(res.error_list), 1)
        # Error type should be SampleSheetError
        self.assertEqual(type(res.error_list[0]), SampleSheetError)

    @patch("iridauploader.parsers.common.get_csv_reader")
    def test_validate_sample_sheet_valid(self, mock_csv_reader):
        """
        Given a valid sample sheet, make sure the response shows as valid
        :param mock_csv_reader:
        :return:
        """
        field_values = (
            "[Data]\n"
            "Sample_Name,Project_ID,File_Forward,File_Reverse\n"
            "my-sample-1,72,file_1.fastq.gz,file_2.fastq.gz\n"
        )

        file_contents_str = (
            "{field_values}"
        ).format(field_values=field_values)

        # converts string as a pseudo file / memory file
        sample_sheet_file = StringIO(file_contents_str)

        # the call to get_csv_reader() inside parse_samples() will return
        # items inside side_effect
        mock_csv_reader.side_effect = [reader(sample_sheet_file)]

        res = validate_sample_sheet(None)

        # This should be a valid sample sheet
        self.assertTrue(res.is_valid())



class ModelValidationError(Exception):
    """An exception raised when errors are encountered while validating the object model.

    """
    def __init__(self, message, object):
        """Initialize a ModelValidationError.

        Args:
            message: a summary message that's causing the error.
            object: the object that has the validation error
        """
        self._message = message
        self._object = object

    @property
    def message(self):
        return self._message

    @property
    def object(self):
        return self._object

    def __str__(self):
        return str(self.message)



import setuptools

# Use the readme file as the long description on PyPi
with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='iridauploader',
    version='0.9.5',
    description='IRIDA uploader: upload NGS data to IRIDA system',
    url='https://github.com/phac-nml/irida-uploader',
    author='Jeffrey Thiessen',
    author_email='jeffrey.thiessen@canada.ca',
    long_description=long_description,
    long_description_content_type="text/markdown",
    # license specified on github
    license='Apache-2.0',
    keywords="IRIDA NGS uploader",
    packages=setuptools.find_packages(include=['iridauploader',
                                               'iridauploader.*',
                                               ]),
    install_requires=['urllib3<2.3.0',
                      'rauth',
                      'requests',
                      'chardet',
                      'appdirs',
                      'cerberus',
                      'argparse',
                      'requests-toolbelt',
                      ],
    extras_require={
        "GUI": ["PyQt5==5.15.2", "PyQt5-stubs==5.14.2.2"],
        "TEST": ["pytest", "coverage"],
        "WINDOWS": ["pynsist"],
    },
    entry_points={
        'console_scripts': [
            'irida-uploader=iridauploader.core.cli:main',
            'irida-uploader-gui=iridauploader.gui.gui:main [GUI]',
            'integration-test=iridauploader.tests_integration.start_integration_tests:main [TEST]'
        ],
    },
    # https://pypi.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
    # urllib3 v2.0 requires >=3.7
    python_requires='>=3.7',

)



"""
This file contains the single entry functions to upload data that are called via the command line and the gui
These functions return exit codes that get passed up the chain to whatever function calls the uploader (i.e. bash)

It also contains the functions for starting/stopping logging to a directory
"""

import logging

from iridauploader import VERSION_NUMBER
import iridauploader.api as api
import iridauploader.config as config
import iridauploader.parsers as parsers
import iridauploader.progress as progress
from iridauploader.model import DirectoryStatus
import os
from . import api_handler, parsing_handler, logger, exit_return, upload_helpers


def upload_run_single_entry(directory, force_upload=False, upload_mode=None, continue_upload=False):
    """
    This function acts as a single point of entry for uploading a directory

    Handles getting a directories run status, and running if conditions are met (valid run, new run or forced upload).

    :param directory: Directory of the sequencing run to upload
    :param force_upload: When set to true, the upload status file will be ignored and file will attempt to be uploaded
    :param upload_mode: String with upload mode to use. When None, default is used.
    :param continue_upload: When set, a PARTIAL status run will be continued from where it left off.
    :return: ExitReturn
    """

    directory_status = parsing_handler.get_run_status(directory)
    parse_as_partial = False

    # Check that directory is writeable, or readonly mode is enabled
    if upload_helpers.directory_has_readonly_conflict(directory_status.directory):
        error_msg = 'Directory cannot be written to. Please check permissions or use readonly mode'
        logging.error(error_msg)
        return exit_error(error_msg)
    # Check if run is New or Delayed, and then do delay logic
    elif (directory_status.status_equals(DirectoryStatus.NEW)
          or directory_status.status_equals(DirectoryStatus.DELAYED)):
        if force_upload:
            logging.debug("Run is skipping delay check via force")
        elif progress.run_is_ready_with_delay(directory_status):
            # Note: This is the "happy path" where upload continues
            logging.debug("Run is ready to upload, continuing")
        else:
            logging.debug("Run is delayed, exiting")
            return exit_success()
    # Check if run is partial, if we are continuing partial runs this block will enable parsing as partial
    # NOTE: its assumed that at most only one of `continue_upload` and `force_upload` are set
    elif directory_status.status_equals(DirectoryStatus.PARTIAL):
        if continue_upload:
            # Happy path for continuing an upload
            logging.info("Continuing upload on a partial run.")
            parse_as_partial = True
        elif force_upload:
            # Note: This is "happy path" 2, where upload starts from beginning with force
            logging.info("Run with status {} is being force uploaded".format(directory_status.status))
        else:
            error_msg = "ERROR! Directory status is PARTIAL. This run can be continued with the --continue_partial" \
                        " argument, or restarted from the beginning with the --force argument."
            logging.error(error_msg)
            return exit_error(error_msg)
    # Check if a run is invalid, an invalid run cannot be uploaded.
    elif directory_status.status_equals(DirectoryStatus.INVALID):
        error_msg = "ERROR! Run in directory {} is invalid. Returned with message: '{}'".format(
            directory_status.directory, directory_status.message)
        logging.error(error_msg)
        return exit_error(error_msg)
    # Check if run is any other status, if force upload is set, continue, otherwise exit
    elif (directory_status.status_equals(DirectoryStatus.ERROR)
          or directory_status.status_equals(DirectoryStatus.COMPLETE)):
        if force_upload:
            # Note: This is "happy path" 2, where upload continues with force
            logging.info("Run with status {} is being force uploaded".format(directory_status.status))
        else:
            error_msg = "ERROR! Run in directory {} is not new. It has either been uploaded, " \
                        "or an upload was attempted with error. " \
                        "Please check the status file 'irida_uploader_status.info' " \
                        "in the run directory for more details. " \
                        "You can bypass this error by uploading with the --force argument.".format(directory)
            logging.error(error_msg)
            return exit_error(error_msg)

    # get default upload mode if None
    if upload_mode is None:
        upload_mode = api_handler.get_default_upload_mode()

    # upload
    return _validate_and_upload(directory_status, upload_mode, parse_as_partial)


def batch_upload_single_entry(batch_directory, force_upload=False, upload_mode=None, continue_upload=False):
    """
    This function acts as a single point of entry for batch uploading run directories

    It uses _validate_and_upload as it function for uploading the individual runs

    A list of runs to be uploaded is generated at start up, and all found runs will be attempted to be uploaded.

    :param batch_directory: Directory containing sequencing run directories to upload
    :param force_upload: When set to true, the upload status file will be ignored and file will attempt to be uploaded
    :param upload_mode: String with upload mode to use. When None, default is used.
    :param continue_upload: When True, continues uploading existing partial runs from where they left off
    :return: ExitReturn
    """
    logging.debug("batch_upload_single_entry:Starting {} with force={}".format(batch_directory, force_upload))

    # get all potential directories to upload
    directory_status_list = parsing_handler.get_run_status_list(batch_directory)
    # list info about directories found
    logging.info("Found {} potential run directories".format(len(directory_status_list)))
    for directory_status in directory_status_list:
        logging.info("DIRECTORY: %s\n"
                     "%30sSTATUS:  %s\n"
                     "%30sDETAILS: %s"
                     % (directory_status.directory, "", directory_status.status, "", directory_status.message))

    # upload_list contains dicts with the following format
    # {'status': DirectoryStatus, 'partial': boolean}
    upload_list = []
    # delayed_list contains DirectoryStatus objects
    delayed_list = []

    for directory_status in directory_status_list:
        logging.info("Analysing directory: {}".format(directory_status.directory))
        # ignore invalid directories
        if directory_status.status_equals(DirectoryStatus.INVALID):
            continue
        # Check if run is New or Delayed, and then do delay logic
        elif (directory_status.status_equals(DirectoryStatus.NEW)
              or directory_status.status_equals(DirectoryStatus.DELAYED)):
            if force_upload:
                logging.debug("BATCH: Run is being added with force")
                upload_list.append({"status": directory_status, "partial": False})
            elif progress.run_is_ready_with_delay(directory_status):
                # Note: This is the "happy path" where upload continues
                logging.debug("BATCH: Run is ready to upload")
                upload_list.append({"status": directory_status, "partial": False})
            else:
                logging.debug("BATCH: Run is delayed")
                delayed_list.append(directory_status)
        # Check if run is partial, if we are continuing partial runs this block will enable parsing as partial
        # NOTE: its assumed that at most only one of `continue_upload` and `force_upload` are set
        elif directory_status.status_equals(DirectoryStatus.PARTIAL):
            if continue_upload:
                # Happy path for continuing an upload
                logging.debug("BATCH: Run is ready to continue upload as partial")
                upload_list.append({"status": directory_status, "partial": True})
            elif force_upload:
                # Note: This is "happy path" 2, where upload continues with force
                logging.debug("BATCH: Partial Run is being added with force")
                upload_list.append({"status": directory_status, "partial": False})
            else:
                logging.debug("BATCH: Partial Run is skipped")
                continue
        # other statuses are error, partial, and complete runs, force upload allows theses
        elif (directory_status.status_equals(DirectoryStatus.ERROR)
              or directory_status.status_equals(DirectoryStatus.COMPLETE)):
            if force_upload:
                logging.debug("BATCH: Run is forced for upload")
                upload_list.append({"status": directory_status, "partial": False})
            else:
                logging.debug("BATCH: Run is skipped")
                continue

    # Display delayed run count to the user
    if len(delayed_list) > 0:
        logging.info("{} run(s) have been delayed.".format(len(delayed_list)))

    # Display runs to upload count to the user
    if force_upload:
        logging.info("Starting upload for all non invalid and non delayed runs. {} run(s) found. "
                     "(Running with --force)".format(len(upload_list)))
    else:
        logging.info("Starting upload for all new runs. {} run(s) found.".format(len(upload_list)))

    # get default upload mode if None
    if upload_mode is None:
        upload_mode = api_handler.get_default_upload_mode()

    # run uploads (including partial), keep track of which directories did not upload
    error_list = []
    for directory_status_dict in upload_list:
        result = _validate_and_upload(directory_status_dict['status'], upload_mode, directory_status_dict['partial'])
        if result.exit_code == exit_return.EXIT_CODE_ERROR:
            error_list.append(directory_status_dict['status'])

    logging.info("Uploads completed with {} error(s)".format(len(error_list)))
    for directory_status in error_list:
        logging.warning("Directory '{}' upload exited with ERROR, check log and status file for details"
                        "".format(directory_status.directory))

    logging.info("Batch upload complete, Exiting!")
    return exit_success()


def _validate_and_upload(directory_status, upload_mode, continue_from_partial):
    """
    This function attempts to upload a single run directory

    Handles parsing and validating the directory for samples
    Sets up the api layer based on config file
    Verifies samples is able to be uploaded (verifies projects exist)
    Initializes objects/routes on IRIDA to accept Samples (creates samples if they don't exist)
    Starts the upload

    :param directory_status: DirectoryStatus object that has directory to try upload
    :param upload_mode: String, mode to use when uploading assemblies
    :param continue_from_partial: when set, already uploaded samples will be skipped, and existing run_id will be used
    :return: ExitReturn
    """
    logging_start_block(directory_status.directory)
    logging.debug("upload_run_single_entry:Starting {}".format(directory_status.directory))
    logging.info("Starting upload for '{}' with continue partial upload = '{}'".format(
        directory_status.directory, str(continue_from_partial)))

    try:
        # Starting upload process: Parse and do offline verification
        sequencing_run = upload_helpers.parse_and_validate(directory_status, continue_from_partial)
        upload_helpers.verify_upload_mode(upload_mode)
        upload_helpers.init_file_status_list_from_sequencing_run(sequencing_run, directory_status)

        # Initialize api, run pre upload preparation and validation
        upload_helpers.initialize_api(directory_status)
        upload_helpers.irida_prep_and_validation(sequencing_run, directory_status)

        # Upload run
        upload_helpers.upload_sequencing_run(sequencing_run, directory_status, upload_mode, continue_from_partial)

    except (progress.exceptions.DirectoryError,
            parsers.exceptions.ValidationError,
            parsers.exceptions.DirectoryError,
            api.exceptions.IridaConnectionError,
            api.exceptions.IridaResourceError,
            api.exceptions.FileError,
            Exception
            ) as e:
        return exit_error(e)

    logging.info("Samples in directory '{}' have finished uploading!".format(directory_status.directory))

    logging_end_block()

    return exit_success()


def exit_error(error):
    """
    Returns an failed run exit code which ends the process when returned
    :return: ExitReturn with EXIT_CODE_ERROR
    """
    logging_end_block()
    return exit_return.ExitReturn(exit_return.EXIT_CODE_ERROR, error)


def exit_success():
    """
    Returns an success run exit code which ends the process when returned
    :return: ExitReturn with EXIT_CODE_SUCCESS
    """
    return exit_return.ExitReturn(exit_return.EXIT_CODE_SUCCESS)


def logging_start_block(directory):
    """
    Logs an information block to the console and file which indicates the start of an upload run.
    Includes the uploader version number set in this module
    :return:
    """
    log_directory = config.read_config_option("log_directory")
    if log_directory:
        run_name = os.path.basename(directory)
        log_directory_run_path = os.path.join(log_directory, run_name)
        if not os.path.isdir(log_directory_run_path):
            os.mkdir(log_directory_run_path)
    else:
        log_directory_run_path = directory
    if config.read_config_option("readonly", bool, False) is False or bool(log_directory):
        logger.add_log_to_directory(log_directory_run_path)
    logging.info("==================================================")
    logging.info("---------------STARTING UPLOAD RUN----------------")
    logging.info("Uploader Version {}".format(VERSION_NUMBER))
    logging.info("Logging to file in: " + logger.get_user_log_dir())
    logging.info("==================================================")


def logging_end_block():
    """
    Logs an block to the console and file that indicates the end of an upload run.
    :return:
    """
    logging.info("==================================================")
    logging.info("----------------ENDING UPLOAD RUN-----------------")
    logging.info("==================================================")
    if config.read_config_option("readonly", bool, False) is False:
        logger.remove_directory_logger()



class IridaConnectionError(Exception):
    """
    This error is thrown when the api cannot connect to the IRIDA api
    Either the server is unreachable or the credentials are invalid

    All calls to the api should expect this error
    """
    pass



import unittest
from unittest.mock import patch
import os

import iridauploader.config as config

path_to_module = os.path.abspath(os.path.dirname(__file__))
if len(path_to_module) == 0:
    path_to_module = '.'


class TestConfig(unittest.TestCase):
    """
    Tests the config module's functionality
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def tearDown(self):
        """
        Deletes files and directories that are used when a test is run
        Sets the module level variables in the config module back to None, so each test is executed from a clean state
        :return:
        """
        # set the config module level variables back to None
        config.config._conf_parser = None
        config.config._user_config_file = None

    @patch("iridauploader.config.config._init_config_parser")
    @patch("iridauploader.config.config._load_config_from_file")
    @patch("iridauploader.config.config._create_new_config_file")
    @patch("os.path.exists")
    def test_basic_setup_all_functions_called(self, mock_path_exists, mock_create_new_config_file,
                                              mock_load_config_from_file, mock_init_config_parser):

        # Set the config file to True so that it thinks there is a config file
        config.set_config_file(True)
        # start up config setup
        config.setup()
        # First init happens
        mock_init_config_parser.assert_called_with()
        # make sure no attempt to create a file happens
        mock_create_new_config_file.assert_not_called()
        # make sure it tries to load from file
        mock_load_config_from_file.assert_called_with()

    @patch("iridauploader.config.config._load_config_from_file")
    @patch("iridauploader.config.config._create_new_config_file")
    @patch("os.path.exists")
    def test_create_new_file_if_none_exist(self, mock_path_exists, mock_create_new_config_file,
                                           mock_load_config_from_file):

        # Set the config file to False so that the function will check if a default file is set
        config.set_config_file(False)
        # When it checks for a default path and override path,
        # return False for bothso it will attempt to create a new file
        mock_path_exists.side_effect = [False, False]
        # try setting up config to trigger creating a new file
        config.setup()
        # make sure the attempt to create a file happens
        mock_create_new_config_file.assert_called_with()
        # It will attempt to use the file that was just created
        mock_load_config_from_file.assert_called_with()

    @patch("iridauploader.config.config._init_config_parser")
    @patch("iridauploader.config.config._load_config_from_file")
    @patch("iridauploader.config.config._create_new_config_file")
    @patch("os.path.exists")
    def test_dont_create_new_file_if_exists(self, mock_path_exists, mock_create_new_config_file,
                                            mock_load_config_from_file, mock_init_config_parser):

        # Set the config file to False so that the function will check if a default file is set
        config.set_config_file(False)
        # When it checks for a default path, return True so it will attempt to use that file
        mock_path_exists.side_effect = [True]
        # If a new file is created, no exit will happen
        config.setup()
        # First init happens
        mock_init_config_parser.assert_called_with()
        # make sure no attempt to create a file happens
        mock_create_new_config_file.assert_not_called()
        # make sure it tries to load from file
        mock_load_config_from_file.assert_called_with()

    @patch("iridauploader.config.config._init_config_parser")
    @patch("iridauploader.config.config._load_config_from_file")
    @patch("iridauploader.config.config._create_new_config_file")
    @patch("os.path.exists")
    def test_dont_create_new_file_if_override_file_exists(self, mock_path_exists, mock_create_new_config_file,
                                                          mock_load_config_from_file, mock_init_config_parser):
        # Set the config file to False so that the function will check if a default file is set
        config.set_config_file(False)
        # When it checks for config files, return True on check for override so it wont attempt to create a new file
        mock_path_exists.side_effect = [False, True]
        # If a new file is created, no exit will happen
        config.setup()
        # First init happens
        mock_init_config_parser.assert_called_with()
        # make sure no attempt to create a file happens
        mock_create_new_config_file.assert_not_called()
        # make sure it tries to load from file
        mock_load_config_from_file.assert_called_with()

    def test_read_config_option(self):
        """
        Test reading from well-defined config file
        :return:
        """
        # set up config
        config.set_config_file(os.path.join(path_to_module, "test_config.conf"))
        config.setup()
        # Test that all the parameters loaded from file are correct
        self.assertEqual(config.read_config_option('client_id'), 'uploader')
        self.assertEqual(config.read_config_option('client_secret'), 'secret')
        self.assertEqual(config.read_config_option('username'), 'admin')
        self.assertEqual(config.read_config_option('password'), 'password1')
        self.assertEqual(config.read_config_option('base_url'), 'http://localhost:8080/irida-latest/api/')
        self.assertEqual(config.read_config_option('parser'), 'miseq')
        self.assertEqual(config.read_config_option('readonly', bool), False)

    def test_read_config_option_case_false(self):
        """
        Test reading from well-defined config file with all lowercase boolean (false)
        :return:
        """
        # set up config
        config.set_config_file(os.path.join(path_to_module, "test_config_case_false.conf"))
        config.setup()
        # Test that all the parameters loaded from file are correct
        self.assertEqual(config.read_config_option('client_id'), 'uploader')
        self.assertEqual(config.read_config_option('client_secret'), 'secret')
        self.assertEqual(config.read_config_option('username'), 'admin')
        self.assertEqual(config.read_config_option('password'), 'password1')
        self.assertEqual(config.read_config_option('base_url'), 'http://localhost:8080/irida-latest/api/')
        self.assertEqual(config.read_config_option('parser'), 'miseq')
        self.assertEqual(config.read_config_option('readonly', bool), False)

    def test_read_config_option_case_true(self):
        """
        Test reading from well-defined config file with strange casing boolean (tRuE)
        :return:
        """
        # set up config
        config.set_config_file(os.path.join(path_to_module, "test_config_case_true.conf"))
        config.setup()
        # Test that all the parameters loaded from file are correct
        self.assertEqual(config.read_config_option('client_id'), 'uploader')
        self.assertEqual(config.read_config_option('client_secret'), 'secret')
        self.assertEqual(config.read_config_option('username'), 'admin')
        self.assertEqual(config.read_config_option('password'), 'password1')
        self.assertEqual(config.read_config_option('base_url'), 'http://localhost:8080/irida-latest/api/')
        self.assertEqual(config.read_config_option('parser'), 'miseq')
        self.assertEqual(config.read_config_option('readonly', bool), True)

    def test_read_config_option_case_error(self):
        """
        Test reading from ill-defined config file with a non-boolean entry
        :return:
        """
        # set up config
        config.set_config_file(os.path.join(path_to_module, "test_config_case_error.conf"))
        config.setup()
        # Test that all the parameters loaded from file are correct
        self.assertEqual(config.read_config_option('client_id'), 'uploader')
        self.assertEqual(config.read_config_option('client_secret'), 'secret')
        self.assertEqual(config.read_config_option('username'), 'admin')
        self.assertEqual(config.read_config_option('password'), 'password1')
        self.assertEqual(config.read_config_option('base_url'), 'http://localhost:8080/irida-latest/api/')
        self.assertEqual(config.read_config_option('parser'), 'miseq')
        with self.assertRaises(NameError) as context:
            self.assertEqual(config.read_config_option('readonly', bool), False)
        self.assertTrue(
            "Config file field 'readonly' expected 'True' or 'False' but instead got 'this_is_not_a_bool'" in
            str(context.exception))
        with self.assertRaises(NameError) as context:
            self.assertEqual(config.read_config_option('http_max_retries', int), False)
        self.assertTrue(
            "Config file field 'http_max_retries' expected int but instead got 'this_is_not_a_int'" in
            str(context.exception))
        with self.assertRaises(NameError) as context:
            self.assertEqual(config.read_config_option('http_backoff_factor', float), False)
        self.assertTrue(
            "Config file field 'http_backoff_factor' expected float but instead got 'this_is_not_a_float'" in
            str(context.exception))

    def test_read_config_option_use_default(self):
        """
        Test reading from ill-defined config file and use default value
        :return:
        """
        # set up config
        config.set_config_file(os.path.join(path_to_module, "test_config_case_error.conf"))
        config.setup()
        # Test that all the parameters loaded from file are correct
        self.assertEqual(config.read_config_option('readonly', bool, False), False)
        self.assertEqual(config.read_config_option('readonly', bool, True), True)
        self.assertEqual(config.read_config_option('http_max_retries', int, 0), 0)
        self.assertEqual(config.read_config_option('http_max_retries', int, 1), 1)
        self.assertEqual(config.read_config_option('http_backoff_factor', float, 0.5), 0.5)
        self.assertEqual(config.read_config_option('http_backoff_factor', float, 0), 0)
        self.assertEqual(config.read_config_option('http_backoff_factor', float, 1), 1)

    def test_set_config_options(self):
        """
        Test writing to config file, make sure writen values are written correctly
        :return:
        """
        # set up config
        config.set_config_file(os.path.join(path_to_module, "test_config.conf"))
        config.setup()
        # Make sure id is initially set to what we expect
        self.assertEqual(config.read_config_option('client_id'), 'uploader')
        self.assertEqual(config.read_config_option('http_max_retries', int), 5)
        self.assertEqual(config.read_config_option('http_backoff_factor', float), 0.5)
        # Set and test to a new id
        config.set_config_options(
            client_id="new_id",
            http_max_retries=2,
            http_backoff_factor=2.5
        )
        self.assertEqual(config.read_config_option('client_id'), "new_id")
        self.assertEqual(config.read_config_option('http_max_retries', int), 2)
        self.assertEqual(config.read_config_option('http_backoff_factor', float), 2.5)



import unittest
from unittest.mock import patch
from os import path
from csv import reader
from io import StringIO

from iridauploader import parsers
import iridauploader.parsers.miniseq.sample_parser as sample_parser
from iridauploader.parsers.exceptions import SampleSheetError, SequenceFileError
from iridauploader.parsers import common
import iridauploader.model as model

path_to_module = path.abspath(path.dirname(__file__))
if len(path_to_module) == 0:
    path_to_module = '.'


class TestParseMetadata(unittest.TestCase):
    """
    Test building meta data from sample sheets with a csv reader
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    @patch("iridauploader.parsers.common.get_csv_reader")
    def test_parse_metadata_paired_valid(self, mock_csv_reader):
        """
        When given a valid directory, ensure valid metadata is built
        paired end reads
        :return:
        """
        h_field_values = (
            "Local Run Manager Analysis Id,4004\n"
            "Experiment Name,Some_Test_Data\n"
            "Date,2015-05-14\n"
            "Workflow,GenerateFastQWorkflow\n"
            "Description,12-34\n"
            "Chemistry,Yes\n"
        )

        reads = (
            "151\n"
            "151\n"
        )

        d_headers = ("Sample_ID,Sample_Name,"
                     "I7_Index_ID,index,I5_Index_ID,index2,Sample_Project")

        d_field_values = ("15-0318-4004,15-0318,N701,TAAGGCGA,S502,CTCTCTAT,203\n"
                          "15-0455-4004,15-0455,N701,TAAGGCGA,S503,TATCCTCT,203\n"
                          "15-0462-4004,15-0462,N701,TAAGGCGA,S505,GTAAGGAG,203\n")

        file_contents_str = (
            "[Header]\n"
            "{h_field_values}\n"
            "[Reads]\n"
            "{reads}\n"
            "[Data]\n"
            "{d_headers}\n"
            "{d_field_values}"
        ).format(h_field_values=h_field_values,
                 reads=reads,
                 d_headers=d_headers,
                 d_field_values=d_field_values)

        # converts string as a pseudo file / memory file
        sample_sheet_file = StringIO(file_contents_str)

        # the call to get_csv_reader() inside parse_samples() will return
        # items inside side_effect
        mock_csv_reader.side_effect = [reader(sample_sheet_file)]

        metadata = sample_parser.parse_metadata(None)
        # The meta data we care about the most
        self.assertEqual(metadata['readLengths'], "151")
        self.assertEqual(metadata['layoutType'], "PAIRED_END")
        # Other meta data should also be here
        self.assertEqual(metadata['localrunmanager'], "4004")
        self.assertEqual(metadata['experimentName'], "Some_Test_Data")
        self.assertEqual(metadata['date'], "2015-05-14")
        self.assertEqual(metadata['workflow'], "GenerateFastQWorkflow")
        self.assertEqual(metadata['description'], "12-34")
        self.assertEqual(metadata['chemistry'], "Yes")

    @patch("iridauploader.parsers.common.get_csv_reader")
    def test_parse_metadata_single_valid(self, mock_csv_reader):
        """
        When given a valid directory, ensure valid metadata is built
        single end reads
        :return:
        """
        h_field_values = (
            "Local Run Manager Analysis Id,4004\n"
            "Experiment Name,Some_Test_Data\n"
            "Date,2015-05-14\n"
            "Workflow,GenerateFastQWorkflow\n"
            "Description,12-34\n"
            "Chemistry,Yes\n"
        )

        reads = (
            "151\n"
        )

        d_headers = ("Sample_ID,Sample_Name,"
                     "I7_Index_ID,index,I5_Index_ID,index2,Sample_Project")

        d_field_values = ("15-0318-4004,15-0318,N701,TAAGGCGA,S502,CTCTCTAT,203\n"
                          "15-0455-4004,15-0455,N701,TAAGGCGA,S503,TATCCTCT,203\n"
                          "15-0462-4004,15-0462,N701,TAAGGCGA,S505,GTAAGGAG,203\n")

        file_contents_str = (
            "[Header]\n"
            "{h_field_values}\n"
            "[Reads]\n"
            "{reads}\n"
            "[Data]\n"
            "{d_headers}\n"
            "{d_field_values}"
        ).format(h_field_values=h_field_values,
                 reads=reads,
                 d_headers=d_headers,
                 d_field_values=d_field_values)

        # converts string as a pseudo file / memory file
        sample_sheet_file = StringIO(file_contents_str)

        # the call to get_csv_reader() inside parse_samples() will return
        # items inside side_effect
        mock_csv_reader.side_effect = [reader(sample_sheet_file)]

        metadata = sample_parser.parse_metadata(None)
        self.assertEqual(metadata['layoutType'], "SINGLE_END")

    def test_parse_metadata(self):
        """
        Testing the parsing meta data with actual files, instead of mocked files
        :return:
        """
        sheet_file = path.join(path_to_module, "fake_ngs_data",
                               "SampleSheet.csv")
        meta_data = sample_parser.parse_metadata(sheet_file)

        correct_metadata = {"readLengths": "151",
                            "workflow": "GenerateFastQWorkflow",
                            "localrunmanager": "4004",
                            "date": "10/15/2013",
                            "chemistry": "Amplicon",
                            "description": "Superbug",
                            "experimentName": '1',
                            "layoutType": "PAIRED_END"}

        self.assertEqual(correct_metadata, meta_data)

    def test_parse_metadata_extra_commas(self):
        """
        Tests parsing meta data when extra commas are in the file
        Testing the parsing meta data with actual files, instead of mocked files
        :return:
        """
        sheet_file = path.join(path_to_module, "testValidSheetTrailingCommas",
                               "SampleSheet.csv")
        meta_data = sample_parser.parse_metadata(sheet_file)

        correct_metadata = {"readLengths": "151",
                            "workflow": "GenerateFastQWorkflow",
                            "localrunmanager": "4004",
                            "date": "10/15/2013",
                            "chemistry": "Amplicon",
                            "description": "Superbug",
                            "experimentName": '1',
                            "layoutType": "PAIRED_END"}

        self.assertEqual(correct_metadata, meta_data)


class TestGetCsvReader(unittest.TestCase):
    """
    Test that the csv reader behaves as expected
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_get_csv_reader_valid_sheet(self):
        """
        Given a valid sample sheet, ensure the parsed sheet matches expected output
        :return:
        """
        sheet_file = path.join(path_to_module, "fake_ngs_data",
                               "SampleSheet.csv")

        lines = common.get_csv_reader(sheet_file)

        correct_lines = [
            ['[Header]'],
            ['Local Run Manager Analysis Id', '4004'],
            ['Experiment Name', '1'],
            ['Date', '10/15/2013'],
            ['Workflow', 'GenerateFastQWorkflow'],
            ['Description', 'Superbug'],
            ['Chemistry', 'Amplicon'],
            [],
            ['[Reads]'],
            ['151'],
            ['151'],
            [],
            ['[Settings]'],
            ['Adapter', 'AAAAGGGGAAAAGGGGAAA'],
            [],
            ['[Data]'],
            ['Sample_ID', 'Sample_Name', 'index', 'I7_Index_ID', 'index2', 'I5_Index_ID',
             'Sample_Project'],
            ['01-1111-4004', '01-1111', 'AAAAAAAA', 'N01', 'TTTTTTTT', 'S01', '6'],
            ['02-2222-4004', '02-2222', 'GGGGGGGG', 'N02', 'CCCCCCCC', 'S02', '6'],
            ['03-3333-4004', '03-3333', 'CCCCCCCC', 'N03', 'GGGGGGGG', 'S03', '6']
        ]

        for line, c_line in zip(lines, correct_lines):
            self.assertEqual(line, c_line)

    def test_get_csv_reader_no_sheet(self):
        """
        Make sure an error is raised if the csv reader is not given a valid sample sheet
        :return:
        """
        sheet_file = path.join(path_to_module, "fake_ngs_data",
                               "Alignment_1")

        with self.assertRaises(SampleSheetError):
            common.get_csv_reader(sheet_file)


class TestValidatePfList(unittest.TestCase):
    """
    Testing the various cases that can appear when validating the pf list
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_valid_lists(self):
        self.assertTrue(sample_parser._validate_pf_list(
            ['01-1111_S1_L001_R1_001.fastq.gz', '01-1111_S1_L001_R2_001.fastq.gz']
        ))
        self.assertTrue(sample_parser._validate_pf_list(
            ['01-1111_S1_L001_R1_001.fastq.gz']
        ))

    def test_invalid_lists(self):
        self.assertFalse(sample_parser._validate_pf_list(
            ['01-1111_S1_L001_R1_001.fastq.gz', '01-1111_S1_L001_R2_001.fastq.gz', '01-1111_S1_L001_R3_001.fastq.gz']
        ))

        self.assertFalse(sample_parser._validate_pf_list([]))

    def test_invalid_file_names(self):
        self.assertFalse(sample_parser._validate_pf_list(
            ['01-1111_S1_L001_R_001.fastq.gz', '01-1111_S1_L001_R_001.fastq.gz']
        ))
        self.assertFalse(sample_parser._validate_pf_list(
            ['01-1111_S1_L001_1_001.fastq.gz', '01-1111_S1_L001_2_001.fastq.gz']
        ))


class TestParseSampleList(unittest.TestCase):
    """
    test parsing the list of samples from a sample sheet
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_valid(self):
        """
        Ensure a a parsed valid directory matches the expected sample list
        :return:
        """
        directory = path.join(path_to_module, "fake_ngs_data")
        sheet_file = path.join(directory, "SampleSheet.csv")
        data_dir = path.join(directory, parsers.miniseq.Parser.get_relative_data_directory())
        data_dir = data_dir.replace("*", "some_dir")
        file_list = parsers.common.get_file_list(data_dir)

        sample = model.Sample(
            "01-1111",
            "",
            1,
            {
                "index": "AAAAAAAA",
                "I7_Index_ID": "N01",
                "sample_project": "6",
                "sequencer_sample_ID": "01-1111-4004",
                "I5_Index_ID": "S01",
                "index2": "TTTTTTTT"
            }
        )

        sequence_file_properties = {
            "sequencer_sample_ID": "01-1111-4004",
            "index": "AAAAAAAA",
            "I7_Index_ID": "N01",
            "I5_Index_ID": "S01",
            "index2": "TTTTTTTT",
            "description": ""
        }

        file_path_1 = path.join(path_to_module,
                                "fake_ngs_data", "Alignment_1", "some_dir", "Fastq", "01-1111_S1_L001_R1_001.fastq.gz")
        file_path_2 = path.join(path_to_module,
                                "fake_ngs_data", "Alignment_1", "some_dir", "Fastq", "01-1111_S1_L001_R2_001.fastq.gz")
        raw_file_list = [file_path_1, file_path_2]

        res = sample_parser.parse_sample_list(sample_sheet_file=sheet_file, run_data_directory=data_dir,
                                              run_data_directory_file_list=file_list, bool_paired_files=True)

        # Check sample is the same
        self.assertEqual(res[0].get_uploadable_dict(), sample.get_uploadable_dict())
        # Check sequencing file is correct
        self.assertEqual(res[0].sequence_file.properties_dict, sequence_file_properties)
        self.assertEqual(res[0].sequence_file.file_list.sort(), raw_file_list.sort())

    def test_not_pf_list(self):
        """
        No Valid files were found with names given in sample sheet
        :return:
        """
        directory = path.join(path_to_module, "ngs_not_pf_list")
        data_dir = path.join(directory, parsers.miniseq.Parser.get_relative_data_directory())
        data_dir = data_dir.replace("*", "some_dir")
        file_list = parsers.common.get_file_list(data_dir)
        file_path = path.join(directory, "SampleSheet.csv")

        with self.assertRaises(SequenceFileError):
            res = sample_parser.parse_sample_list(sample_sheet_file=file_path, run_data_directory=data_dir,
                                                  run_data_directory_file_list=file_list, bool_paired_files=True)

    def test_not_valid_pf_list(self):
        """
        The file list in the sample sheet is invalid
        :return:
        """
        directory = path.join(path_to_module, "ngs_not_valid_pf_list")
        data_dir = path.join(directory, parsers.miniseq.Parser.get_relative_data_directory())
        data_dir = data_dir.replace("*", "some_dir")
        file_list = parsers.common.get_file_list(data_dir)
        file_path = path.join(directory, "SampleSheet.csv")

        with self.assertRaises(SequenceFileError):
            res = sample_parser.parse_sample_list(sample_sheet_file=file_path, run_data_directory=data_dir,
                                                  run_data_directory_file_list=file_list, bool_paired_files=True)

    def test_space_in_sample_name(self):
        directory = path.join(path_to_module, "ngs_space_in_sample_name")
        data_dir = path.join(directory, parsers.miniseq.Parser.get_relative_data_directory())
        data_dir = data_dir.replace("*", "some_dir")
        file_list = parsers.common.get_file_list(data_dir)
        file_path = path.join(directory, "SampleSheet.csv")

        # Just making sure this doesn't throw an error
        sample_parser.parse_sample_list(sample_sheet_file=file_path, run_data_directory=data_dir,
                                        run_data_directory_file_list=file_list, bool_paired_files=False)


class TestParseSamples(unittest.TestCase):
    """
    Test validity or invalidity of parsed samples
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_parse_samples_valid(self):
        """
        Ensure a a parsed valid directory matches the expected samples
        :return:
        """
        sheet_file = path.join(path_to_module, "fake_ngs_data",
                               "SampleSheet.csv")

        sample1 = model.Sample(
            "01-1111",
            "",
            1,
            {
                "index": "AAAAAAAA",
                "I7_Index_ID": "N01",
                "sample_project": "6",
                "sequencer_sample_ID": "01-1111-4004",
                "I5_Index_ID": "S01",
                "index2": "TTTTTTTT"
            }
        )

        sample2 = model.Sample(
            "02-2222",
            "",
            2,
            {
                "index": "GGGGGGGG",
                "I7_Index_ID": "N02",
                "sample_project": "6",
                "sequencer_sample_ID": "02-2222-4004",
                "I5_Index_ID": "S02",
                "index2": "CCCCCCCC"
            }
        )

        sample3 = model.Sample(
            "03-3333",
            "",
            3,
            {
                "index": "CCCCCCCC",
                "I7_Index_ID": "N03",
                "sample_project": "6",
                "sequencer_sample_ID": "03-3333-4004",
                "I5_Index_ID": "S03",
                "index2": "GGGGGGGG"
            }
        )

        correct_samples = [sample1, sample2, sample3]

        res = sample_parser._parse_samples(sheet_file)
        for r_sample, c_sample in zip(res, correct_samples):
            self.assertEqual(r_sample.get_uploadable_dict(), c_sample.get_uploadable_dict())


class TestParseOutSequenceFile(unittest.TestCase):
    """
    test building the sample object
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_parse_out_sequence_file(self):
        """
        Tests that parse out sequence file correctly filters sample related data from the extra params dict
        And ensures that the uploadable dict correctly includes all the needed data after removal
        :return:
        """
        sample = model.Sample(
            "03-3333",
            "Deadly bug",
            None,
            {
                "index": "CCCCCCCC",
                "I7_Index_ID": "N03",
                "sample_project": "6",
                "sequencer_sample_ID": "03-3333-4004",
                "I5_Index_ID": "S03",
                "index2": "GGGGGGGG"

            }
        )

        uploadable_dict = {'index': 'CCCCCCCC',
                           'I7_Index_ID': 'N03',
                           'sampleName': '03-3333',
                           'sample_project': '6',
                           'sequencer_sample_ID': '03-3333-4004',
                           'I5_Index_ID': 'S03',
                           'index2': 'GGGGGGGG',
                           'description': 'Deadly bug'}

        sequence_file_dict = {'index': 'CCCCCCCC',
                              'I7_Index_ID': 'N03',
                              'sequencer_sample_ID': '03-3333-4004',
                              'I5_Index_ID': 'S03',
                              'index2': 'GGGGGGGG',
                              'description': 'Deadly bug'}

        res = sample_parser._parse_out_sequence_file(sample)

        self.assertEqual(sample.get_uploadable_dict(), uploadable_dict)
        self.assertEqual(res, sequence_file_dict)


class TestBuildISeqRun(unittest.TestCase):
    """
    Test building the an iseq run from a sample sheet with a csv reader
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_build_valid_with_description_field(self):
        """
        When given a valid directory, ensure a valid SequencingRun is built with Projects, Samples, ect
        :return:
        """
        directory = path.join(path_to_module, "iseq_with_desc_field")
        sheet_file = path.join(directory, "SampleSheet.csv")
        meta_data = sample_parser.parse_metadata(sheet_file)
        data_dir = path.join(directory, parsers.miniseq.Parser.get_relative_data_directory())
        data_dir = data_dir.replace("*", "some_dir")
        file_list = parsers.common.get_file_list(data_dir)

        sample_list = sample_parser.parse_sample_list(sample_sheet_file=sheet_file,
                                                      run_data_directory=data_dir,
                                                      run_data_directory_file_list=file_list,
                                                      bool_paired_files=True)
        sequence_run_type = 'miniseq'

        sequencing_run = parsers.common.build_sequencing_run_from_samples(sample_list, meta_data, sequence_run_type)

        # Returns a SequencingRun
        self.assertEqual(type(sequencing_run), model.SequencingRun)
        # Includes a single project
        self.assertEqual(len(sequencing_run.project_list), 1)
        # is of type Project
        self.assertEqual(type(sequencing_run.project_list[0]), model.Project)
        # Project has 3 samples
        self.assertEqual(len(sequencing_run.project_list[0].sample_list), 3)
        # samples are of type Sample
        self.assertEqual(type(sequencing_run.project_list[0].sample_list[0]), model.Sample)
        # samples have correct description
        self.assertEqual(sequencing_run.project_list[0].sample_list[0].description, "desc1")
        # samples have SequenceFile
        self.assertEqual(type(sequencing_run.project_list[0].sample_list[0].sequence_file), model.SequenceFile)



"""
This file holds helper functions that assist upload.py

Each function handles it's own try/except block, and updates the directory status
as they progress through their respective tasks
"""

import logging
import os

from pprint import pformat

from iridauploader import api
import iridauploader.config as config
import iridauploader.parsers as parsers
import iridauploader.progress as progress
from iridauploader.model import DirectoryStatus

from . import api_handler, parsing_handler


def _set_and_write_directory_status(directory_status, status, message=None):
    """
    Given a DirectoryStatus object, sets the status and message, and then writes to the directory status directory

    :param directory_status: DirectoryStatus object
    :param status: a valid DirectoryStatus status
    :param message: string
    :return:
    """
    # If a run is being set to error, but it already has a sequencing id (started upload),
    # it is a partial run that could be resumed.
    if status == DirectoryStatus.ERROR and directory_status.run_id is not None:
        status = DirectoryStatus.PARTIAL

    try:
        directory_status.status = status
        directory_status.message = message
        progress.write_directory_status(directory_status)
    except progress.exceptions.DirectoryError as e:
        logging.error("ERROR! Error while trying to write status file to directory {} with error message: {}"
                      "".format(e.directory, e.message))
        raise e


# ****************************************************************
# upload_single_entry / batch_upload_single_entry helper functions
# ****************************************************************

def directory_has_readonly_conflict(directory):
    """
    Returns True if directory is not writable and readonly is not True
    Else returns False
    :param directory:
    :return: boolean
    """
    if (config.read_config_option("readonly", bool, False) is False and not os.access(directory, os.W_OK)):
        return True
    else:
        return False


def set_run_delayed(directory_status):
    """
    Set the directory_status's status to be DELAYED
    :param directory_status:
    :return:
    """
    _set_and_write_directory_status(directory_status, DirectoryStatus.DELAYED)


# *************************************
# _validate_and_upload helper functions
# *************************************


def parse_and_validate(directory_status, parse_as_partial):
    """
    Do the parsing, and offline validation
    :param directory_status: DirectoryStatus object
    :param parse_as_partial: sequencing_run will not include any samples that have already been uploaded
    :return:
    """

    try:
        sequencing_run = parsing_handler.parse_and_validate(directory_status.directory)
    except parsers.exceptions.DirectoryError as e:
        # Directory was not valid for some reason
        full_error = "ERROR! An error occurred with directory '{}', with message: {}".format(e.directory, e.message)
        logging.error(full_error)
        logging.info("Samples not uploaded!")
        _set_and_write_directory_status(directory_status, DirectoryStatus.ERROR, full_error)
        raise e
    except parsers.exceptions.ValidationError as e:
        # Sequencing Run / SampleSheet was not valid for some reason
        error_msg = "ERROR! Errors occurred during validation with message: {}".format(e.message)
        logging.error(error_msg)
        error_list_msg = "Error list: " + pformat(e.validation_result.error_list)
        logging.error(error_list_msg)
        logging.info("Samples not uploaded!")
        full_error = error_msg + ", " + error_list_msg
        _set_and_write_directory_status(directory_status, DirectoryStatus.ERROR, full_error)
        raise e
    # When continuing a partial run, filter out already uploaded samples
    if parse_as_partial:
        sequencing_run = set_uploaded_samples_to_skip(sequencing_run, directory_status.get_sample_status_list())

    return sequencing_run


def set_uploaded_samples_to_skip(sequencing_run, sample_status_list):
    """
    Given a complete sequencing run, remove all samples that have been marked as uploaded in the sample status list
    :param sequencing_run: sequencing run to filter
    :param sample_status_list: full sample list to check against
    :return: sequencing_run
    """
    # This block looks worse than it is, worst case is O(n^2)
    # There are "cleaner" but less readable ways to do this in the same runtime, but readability > reduced line count
    for project in sequencing_run.project_list:
        for sample in project.sample_list:
            for sample_status in sample_status_list:
                if (sample_status.uploaded is True
                        and sample_status.sample_name == sample.sample_name
                        and str(sample_status.project_id) == str(project.id)):
                    # If a sample has already been uploaded, skip it
                    sample.skip = True

    return sequencing_run


def verify_upload_mode(upload_mode):
    """
    Verifies upload mode string given is a valid upload mode
    :param upload_mode: String
    :return: None, throws Exception if invalid
    """
    valid_upload_mode_list = api_handler.get_upload_modes()
    if upload_mode not in valid_upload_mode_list:
        e = Exception("Upload mode '{}' is not valid, upload mode must be one of {}".format(
            upload_mode,
            valid_upload_mode_list
        ))
        logging.error(e)
        raise e


def init_file_status_list_from_sequencing_run(sequencing_run, directory_status):
    """
    Update status file with sample list and write
    :param sequencing_run:
    :param directory_status:
    :return: None
    """
    directory_status.init_file_status_list_from_sequencing_run(sequencing_run)


def initialize_api(directory_status):
    """
    Setup api handler for first use
    :param directory_status
    :return: None
    """
    try:
        api_handler.initialize_api_from_config()
    except api.exceptions.IridaConnectionError as e:
        logging.error("ERROR! Could not initialize irida api.")
        logging.error("Errors: " + pformat(e.args))
        logging.info("Samples not uploaded!")
        full_error = "ERROR! Could not initialize irida api. Errors: " + pformat(e.args)
        _set_and_write_directory_status(directory_status, DirectoryStatus.ERROR, full_error)
        raise e
    logging.info("*** Connected ***")


def irida_prep_and_validation(sequencing_run, directory_status):
    """
    Prepare sequencing run for upload to IRIDA and run validation on against the sequencing_run
    :param sequencing_run:
    :param directory_status
    :return: ValidationResult object
    """
    logging.info("*** Verifying run (online validation) ***")
    try:
        validation_result = api_handler.prepare_and_validate_for_upload(sequencing_run)
    except api.exceptions.IridaConnectionError as e:
        logging.error("Lost connection to Irida")
        logging.error("Errors: " + pformat(e.args))
        full_error = "Lost connection to Irida. Errors: " + pformat(e.args)
        _set_and_write_directory_status(directory_status, DirectoryStatus.ERROR, full_error)
        raise e

    if not validation_result.is_valid():
        logging.error("Sequencing run can not be uploaded")
        logging.error("Sequencing run can not be uploaded. Encountered {} errors"
                      "".format(validation_result.error_count()))
        logging.error("Errors: " + pformat(validation_result.error_list))
        full_error = "Sequencing run can not be uploaded, Errors: " + pformat(validation_result.error_list)
        _set_and_write_directory_status(directory_status, DirectoryStatus.ERROR, full_error)
        raise Exception(full_error)
    logging.info("*** Run Verified ***")


def upload_sequencing_run(sequencing_run, directory_status, upload_mode, upload_from_partial=False):
    """
    Starts the actual upload of the sequencing run
    :param sequencing_run:
    :param directory_status:
    :param upload_mode:
    :param upload_from_partial: Default False, when continuing from a partial run, we can reuse the the run_id
    :return: None
    """
    logging.info("*** Starting Upload ***")
    try:
        # If continuing a partial run, use the existing run_id
        if upload_from_partial:
            run_id = directory_status.run_id
        else:
            run_id = None
        # Upload
        api_handler.upload_sequencing_run(
            sequencing_run=sequencing_run,
            directory_status=directory_status,
            upload_mode=upload_mode,
            run_id=run_id
        )
    except api.exceptions.IridaConnectionError as e:
        logging.error("Lost connection to Irida")
        logging.error("Errors: " + pformat(e.args))
        full_error = "Lost connection to Irida. Errors: " + pformat(e.args)
        _set_and_write_directory_status(directory_status, DirectoryStatus.ERROR, full_error)
        raise e
    except api.exceptions.IridaResourceError as e:
        logging.error("Could not access IRIDA resource")
        logging.error("Errors: " + pformat(e.args))
        full_error = "Could not access IRIDA resource Errors: " + pformat(e.args)
        _set_and_write_directory_status(directory_status, DirectoryStatus.ERROR, full_error)
        raise e
    except api.exceptions.FileError as e:
        logging.error("Could not upload file to IRIDA")
        logging.error("Errors: " + pformat(e.args))
        full_error = "Could not upload file to IRIDA. Errors: " + pformat(e.args)
        _set_and_write_directory_status(directory_status, DirectoryStatus.ERROR, full_error)
        raise e

    _set_and_write_directory_status(directory_status, DirectoryStatus.COMPLETE)
    logging.info("*** Upload Complete ***")
